#! /usr/bin/env python3
# -*- coding: utf-8 -*-
"""Commandline API gateway for laskea."""
import sys
from typing import List

import requests_cache
import typer

import laskea
import laskea.config as cfg
import laskea.laskea as fill
from laskea import env

app = typer.Typer(
    add_completion=False,
    context_settings={'help_option_names': ['-h', '--help']},
    no_args_is_help=True,
)


@app.callback(invoke_without_command=True)
def callback(
    version: bool = typer.Option(
        False,
        '-V',
        '--version',
        help='Display the laskea version and exit',
        is_eager=True,
    )
) -> None:
    """
    Calculate (Finnish: laskea) some parts.
    """
    if version:
        typer.echo(f'{laskea.APP_NAME} version {laskea.__version__}')
        raise typer.Exit()


@app.command('template')
def app_template() -> int:
    """
    Write a template of a well-formed JSON configuration to standard out and exit

    The strategy for looking up configurations is to start at the current working
    directory trying to read a file with the name `.laskea.json` else try to read
    same named file in the user folder (home).

    In case an explicit path is given to the config option of commands that offer
    it, only that path is considered.
    """
    sys.stdout.write(cfg.generate_template())
    return sys.exit(0)


@app.command('report')
def report(
    shallow: bool = typer.Option(
        False,
        '-s',
        '--shallow',
        help='Shallow reporting - no setuptools required (default is False)',
    ),
) -> int:
    """Output either text options for the user to report her env or the report of the environment for support."""
    sys.stdout.write(env.report(shallow))
    return sys.exit(0)


@app.command('update')
def update(  # noqa
    source: List[str],
    inp: str = typer.Option(
        '',
        '-i',
        '--input',
        help='Path to input file',
        metavar='<sourcepath>',
    ),
    conf: str = typer.Option(
        '',
        '-c',
        '--config',
        help=f'Path to config file (default is $HOME/{laskea.DEFAULT_CONFIG_NAME})',
        metavar='<configpath>',
    ),
    verify: bool = typer.Option(
        False,
        '-n',
        '--dry-run',
        help='Dry run (default is False)',
    ),
    verbose: bool = typer.Option(
        False,
        '-v',
        '--verbose',
        help='Verbose output (default is False)',
    ),
    quiet: bool = typer.Option(
        False,
        '-q',
        '--quiet',
        help='Minimal output (default is False)',
    ),
    strict: bool = typer.Option(
        False,
        '-s',
        '--strict',
        help='Ouput noisy warnings on console and in the processed document (default is False)',
    ),
    expires: int = typer.Option(
        180,
        '-x',
        '--cache-expiry-seconds',
        help='Request cache expiry in seconds (default is 180)',
    ),
) -> int:
    """
    Fill in some parts of the input document.

    You can set some options per evironment variables:

    \b
    * LASKEA_USER='remote-user'
    * LASKEA_TOKEN='remote-secret'
    * LASKEA_BASE_URL='https://remote-jira-instance.example.com/'
    * LASKEA_CACHE_EXPIRY_SECONDS=180
    * LASKEA_COL_FIELDS: '["Key", "Summary", "Custom Field Name"]'
    * LASKEA_COL_MAPS='{"key": ["key", "key"], "summary": ["summary", "fields.summary"],
      "custom field name": ["customfield_123", "fields.customfield_123"]}'
    * LASKEA_JOIN_STRING=' <br>'
    * LASKEA_LF_ONLY='AnythingTruthy'
    * LASKEA_IS_CLOUD='WhenNotConnectingToJiraServerButJiraCloud'
    * LASKEA_MARKERS='[[[fill ]]] [[[end]]]'
    * LASKEA_DEBUG='AnythingTruthy'
    * LASKEA_VERBOSE='AnythingTruthy'
    * LASKEA_STRICT='AnythingTruthy'

    The quiet option (if given) disables any conflicting verbosity setting.
    """
    command = 'update'
    transaction_mode = 'commit' if not verify else 'dry-run'
    if quiet:
        laskea.QUIET = True
        laskea.DEBUG = False
        laskea.VERBOSE = False
    elif verbose:
        laskea.VERBOSE = True

    if strict:
        laskea.STRICT = True

    if transaction_mode == 'dry-run':
        laskea.DRY_RUN = True

    requests_cache.install_cache(cache_name='.laskea_cache', backend='sqlite', expire_after=expires)
    laskea.CACHE_EXPIRY_SECONDS = expires
    options = {
        'quiet': quiet,
        'strict': strict,
        'verbose': verbose,
    }
    cfg.process(conf, options)

    paths = (inp,) if inp else tuple(source)
    return sys.exit(fill.process(command, transaction_mode, paths, options))


@app.command('version')
def app_version() -> None:
    """
    Display the laskea version and exit.
    """
    callback(True)



import pathlib

__all__ = ['sbom_sha256']

ENCODING = 'utf-8'
TARGET = """\
The [SBOM in CycloneDX v1.4 JSON format](https://github.com/sthagen/laskea/blob/default/sbom.json) with\
 SHA256 checksum ([$hash_8$ ...](https://raw.githubusercontent.com/sthagen/laskea/default/sbom.json.sha256\
 "sha256:$hash_full$")).\
"""


def sbom_sha256():
    """Fill in the data."""
    with open(pathlib.Path('sbom.json.sha256'), 'rt', encoding=ENCODING) as handle:
        hash_full = handle.read().strip()
    hash_8 = hash_full[:8]
    print(TARGET.replace('$hash_8$', hash_8).replace('$hash_full$', hash_full))



import pytest

import laskea.laskea as fill


def test_process_wrong_command(capsys):
    err_msg = 'Usage: laskea update [--help] [-v] [-c config-path] [-n] [-i] source-files*md [other.md]'
    with pytest.raises(SystemExit):
        fill.process('wrong', 'ignore', ('ignore',), {})
    out, err = capsys.readouterr()
    assert not err
    assert err_msg in out


def test_process_empty_paths(capsys):
    err_msg = 'Usage: laskea update [--help] [-v] [-c config-path] [-n] [-i] source-files*md [other.md]'
    with pytest.raises(SystemExit):
        fill.process('update', 'ignore', tuple(), {})
    out, err = capsys.readouterr()
    assert not err
    assert err_msg in out



import pytest

import laskea
import laskea.api.jira as impl
import laskea.embed as emb

URL_FIXTURE = 'https://remote-jira-instance.example.com'
HEADING_FIXTURE = {
    'rows': [
        {
            'key': 'ABC-42',
            'summary': 'Some issue to show off the headings',
        }
    ]
}
HEADING_PAYLOAD_POSTFIX = (
    f' [{HEADING_FIXTURE["rows"][0]["key"]}]('
    f'{URL_FIXTURE}/browse/{HEADING_FIXTURE["rows"][0]["key"]})'
    f' - {HEADING_FIXTURE["rows"][0]["summary"]}\n'
)
HEADING_WARN_TOO_MANY_FIXTURE = {
    'rows': [
        {
            'key': 'ABC-42',
            'summary': 'Some issue to show off the headings',
        },
        {
            'key': 'ABC-1001',
            'summary': 'Some additional issue to show off the headings warning',
        },
    ]
}
HEADING_WARN_TOO_MANY_FIXTURE_COUNT = len(HEADING_WARN_TOO_MANY_FIXTURE['rows'])
HEADING_WARN_TOO_FEW_FIXTURE = {'rows': []}

UO_LIST_FIXTURE = {
    'rows': [
        {
            'key': 'ABC-42',
            'summary': 'First issue to show off the ordered and unordered lists',
        },
        {
            'key': 'ABC-1001',
            'summary': 'Second issue to show off the ordered and unordered lists',
        },
    ]
}
UO_LIST_PAYLOAD_POSTFIXES = tuple(
    f' [{row["key"]}]({URL_FIXTURE}/browse/{row["key"]})' f' - {row["summary"]}' for row in UO_LIST_FIXTURE['rows']
)

D_LIST_FIXTURE = {
    'rows': [
        {
            'key': 'ABC-42',
            'summary': 'First issue to show off the definition lists',
        },
        {
            'key': 'ABC-1001',
            'summary': 'Second issue to show off the definition lists',
        },
    ]
}
D_LIST_PAYLOADS = tuple(
    f'[{row["key"]}]({URL_FIXTURE}/browse/{row["key"]})\n' f':{row["summary"]}\n' for row in D_LIST_FIXTURE['rows']
)

TABLE_FIXTURE = {
    'rows': [
        {
            'key': 'ABC-42',
            'summary': 'First issue to show off the tables',
        },
        {
            'key': 'ABC-1001',
            'summary': 'Second issue to show off the tables',
        },
    ]
}
TABLE_FIXTURE_PAYLOADS = (
    '| key                                                                  | summary                             |',
    '|:---------------------------------------------------------------------|:------------------------------------|',
    '| [ABC-42](https://remote-jira-instance.example.com/browse/ABC-42)     | First issue to show off the tables  |',
    '| [ABC-1001](https://remote-jira-instance.example.com/browse/ABC-1001) | Second issue to show off the tables |',
    '',
    '2 issues',
)


@pytest.mark.parametrize('level', [lv for lv in range(1, 6 + 1)])
def test_embed_headings(level, capsys):
    impl.BASE_URL = URL_FIXTURE
    hx = [None, emb.h1, emb.h2, emb.h3, emb.h4, emb.h5, emb.h6]
    token = '#' * level
    assert hx[level](query_text='', data=HEADING_FIXTURE) is None
    out, err = capsys.readouterr()
    assert not err
    assert out.strip() == f'{token}{HEADING_PAYLOAD_POSTFIX}'.strip()


@pytest.mark.parametrize('kind,marker', [('ol', '1.'), ('ul', '-')])
def test_embed_uo_lists(kind, marker, capsys):
    impl.BASE_URL = URL_FIXTURE
    lx = {'ol': emb.ol, 'ul': emb.ul}
    expected = '\n'.join(f'{marker}{UO_LIST_PAYLOAD_POSTFIXES[slot]}' for slot in (0, 1))
    assert lx[kind](query_text='', data=UO_LIST_FIXTURE) is None
    out, err = capsys.readouterr()
    assert not err
    assert out.strip() == expected


def test_embed_d_list(capsys):
    impl.BASE_URL = URL_FIXTURE
    expected = '\n'.join(D_LIST_PAYLOADS) + '\n' + '\n'
    assert emb.dl(query_text='', data=D_LIST_FIXTURE) is None
    out, err = capsys.readouterr()
    assert not err
    assert out == expected


@pytest.mark.parametrize('level', [lv for lv in range(1, 6 + 1)])
def test_embed_headings_too_few_strict(level, capsys):
    impl.BASE_URL = URL_FIXTURE
    strictness = laskea.STRICT
    laskea.STRICT = True
    hx = [None, emb.h1, emb.h2, emb.h3, emb.h4, emb.h5, emb.h6]
    query_text = ''
    assert hx[level](query_text=query_text, data=HEADING_WARN_TOO_FEW_FIXTURE) is None
    out, err = capsys.readouterr()
    assert not err  # f'WARNING: received 0 results instead of 1 for JQL ({query_text}) and h{level}' in err
    assert out == f'WARNING: received 0 results instead of 1 for JQL () and h{level}\n'
    laskea.STRICT = strictness


@pytest.mark.parametrize('level', [lv for lv in range(1, 6 + 1)])
def test_embed_headings_too_many_strict(level, capsys):
    impl.BASE_URL = URL_FIXTURE
    strictness = laskea.STRICT
    laskea.STRICT = True
    hx = [None, emb.h1, emb.h2, emb.h3, emb.h4, emb.h5, emb.h6]
    query_text = ''
    # mis_count = HEADING_WARN_TOO_MANY_FIXTURE_COUNT
    assert hx[level](query_text=query_text, data=HEADING_WARN_TOO_MANY_FIXTURE) is None
    out, err = capsys.readouterr()
    assert not err  # f'WARNING: received {mis_count} results instead of 1 for JQL ({query_text}) and h{level}' in err
    assert out
    laskea.STRICT = strictness


def test_embed_table(capsys):
    impl.BASE_URL = URL_FIXTURE
    expected = '\n'.join(TABLE_FIXTURE_PAYLOADS)
    assert emb.table(query_text='', data=TABLE_FIXTURE) is None
    out, err = capsys.readouterr()
    assert not err
    assert out.strip() == expected


def test_embed_table_no_result_strict(capsys):
    impl.BASE_URL = URL_FIXTURE
    strictness = laskea.STRICT
    laskea.STRICT = True
    query_text = ''
    expected = f'WARNING: received 0 results for JQL ({query_text}) and table'
    assert emb.table(query_text=query_text, data=HEADING_WARN_TOO_FEW_FIXTURE) is None
    out, err = capsys.readouterr()
    assert not err  # == 'WARNING: received 0 results for JQL () and table\n'
    assert out.strip() == expected
    laskea.STRICT = strictness


def test_embed_table_no_result_non_strict(capsys):
    impl.BASE_URL = URL_FIXTURE
    strictness = laskea.STRICT
    laskea.STRICT = False
    query_text = ''
    expected = ''
    assert emb.table(query_text=query_text, data=HEADING_WARN_TOO_FEW_FIXTURE) is None
    out, err = capsys.readouterr()
    assert not err
    assert out.strip() == expected
    laskea.STRICT = strictness


@pytest.mark.parametrize('level', [lv for lv in range(1, 6 + 1)])
def test_embed_headings_too_few_non_strict(level, capsys):
    impl.BASE_URL = URL_FIXTURE
    strictness = laskea.STRICT
    laskea.STRICT = False
    hx = [None, emb.h1, emb.h2, emb.h3, emb.h4, emb.h5, emb.h6]
    query_text = ''
    assert hx[level](query_text=query_text, data=HEADING_WARN_TOO_FEW_FIXTURE) is None
    out, err = capsys.readouterr()
    assert not err
    assert not out.strip()
    laskea.STRICT = strictness


@pytest.mark.parametrize('level', [lv for lv in range(1, 6 + 1)])
def test_embed_headings_too_many_non_strict(level, capsys):
    impl.BASE_URL = URL_FIXTURE
    strictness = laskea.STRICT
    laskea.STRICT = False
    hx = [None, emb.h1, emb.h2, emb.h3, emb.h4, emb.h5, emb.h6]
    query_text = ''
    assert hx[level](query_text=query_text, data=HEADING_WARN_TOO_MANY_FIXTURE) is None
    out, err = capsys.readouterr()
    assert not err
    assert not out.strip()
    laskea.STRICT = strictness



import json
import pathlib

import pytest

import laskea.api.jira as impl

URL_FIXTURE = ''
HEADING_FIXTURE = {
    'rows': [
        {
            'key': 'ABC-42',
            'summary': 'Some issue to show off the headings',
        }
    ]
}
HEADING_PAYLOAD_POSTFIX = (
    f' [{HEADING_FIXTURE["rows"][0]["key"]}]({URL_FIXTURE}/browse/{HEADING_FIXTURE["rows"][0]["key"]})'
    f' - {HEADING_FIXTURE["rows"][0]["summary"]}'
)

UO_LIST_FIXTURE = {
    'rows': [
        {
            'key': 'ABC-42',
            'summary': 'First issue to show off the ordered and unordered lists',
        },
        {
            'key': 'ABC-1001',
            'summary': 'Second issue to show off the ordered and unordered lists',
        },
    ]
}
UO_LIST_PAYLOAD_POSTFIXES = tuple(
    f' [{row["key"]}]({URL_FIXTURE}/browse/{row["key"]}) - {row["summary"]}' for row in UO_LIST_FIXTURE['rows']
)

D_LIST_FIXTURE = {
    'rows': [
        {
            'key': 'ABC-42',
            'summary': 'First issue to show off the definition lists',
        },
        {
            'key': 'ABC-1001',
            'summary': 'Second issue to show off the definition lists',
        },
    ]
}
D_LIST_PAYLOADS = tuple(
    f'[{row["key"]}]({URL_FIXTURE}/browse/{row["key"]})\n:{row["summary"]}\n' for row in D_LIST_FIXTURE['rows']
)

P_C_L_FIXTURE_PATH = pathlib.Path('test', 'fixtures', 'basic', 'p_c_jira.json')


@pytest.mark.parametrize('level', [lv for lv in range(1, 6 + 1)])
def test_impl_headings(level):
    impl.BASE_URL = URL_FIXTURE
    hd = impl.markdown_heading(impl.Jira(''), jql_text='', column_fields=tuple(), level=level, data=HEADING_FIXTURE)
    token = '#' * level
    assert hd == f'{token}{HEADING_PAYLOAD_POSTFIX}'


@pytest.mark.parametrize('kind,marker', [('ol', '1.'), ('ul', '-')])
def test_impl_uo_lists(kind, marker):
    impl.BASE_URL = URL_FIXTURE
    text = impl.markdown_list(impl.Jira(''), jql_text='', column_fields=tuple(), list_type=kind, data=UO_LIST_FIXTURE)
    items = text.strip().split('\n')
    for slot in (0, 1):
        assert items[slot] == f'{marker}{UO_LIST_PAYLOAD_POSTFIXES[slot]}'


def test_impl_d_list():
    impl.BASE_URL = URL_FIXTURE
    text = impl.markdown_list(impl.Jira(''), jql_text='', column_fields=tuple(), list_type='dl', data=D_LIST_FIXTURE)
    assert text == '\n'.join(D_LIST_PAYLOADS) + '\n'


def test_impl_login_no_user():
    impl.BASE_USER = ''
    message = 'User, Token, and URL are all required for login.'
    with pytest.raises(ValueError, match=message):
        _ = impl.login(url='does-not-help-as-user-is-missing')


def test_impl_login_no_token():
    impl.BASE_TOKEN = ''
    message = 'User, Token, and URL are all required for login.'
    with pytest.raises(ValueError, match=message):
        _ = impl.login(user='not-relevant-as-token-is-missing')


def test_impl_login_no_url():
    impl.BASE_URL = ''
    message = 'User, Token, and URL are all required for login.'
    with pytest.raises(ValueError, match=message):
        _ = impl.login(user='not-relevant', token='as-url-is-missing')


def test_impl_login_no_url_server():
    impl.BASE_URL = ''
    message = 'User, Token, and URL are all required for login.'
    with pytest.raises(ValueError, match=message):
        _ = impl.login(user='not-relevant', token='as-url-is-missing', is_cloud=False)


@pytest.mark.parametrize('is_cloud', [True, False])
def test_impl_login_trying_but_bad_non_url(is_cloud, capsys):
    impl.BASE_URL = ''
    impl.BASE_IS_CLOUD = is_cloud
    _ = impl.login(user='not-relevant', token='as-url-is-leading-nowhere', url='/dev/null', is_cloud=is_cloud)
    out, err = capsys.readouterr()
    assert not err
    assert not out


def test_impl_query():
    impl.BASE_COL_FIELDS = ('key', 'summary')
    result = impl.query(impl.Jira(''), jql_text='', column_fields=tuple())
    assert result == {'jql_text': '', 'error': 'Empty JIRA Query Language text detected'}


def test_parent_children_sections():
    with open(P_C_L_FIXTURE_PATH, 'rt', encoding='utf-8') as handle:
        data = json.load(handle)

    doc = impl.parent_children_sections(impl.Jira(''), 'parent_jql', 'children_jql', 'Test Plan', 'Test Case', data)
    assert '## First summary' in doc



"""Calculate (Finnish: laskea) some parts."""
import sys
from typing import Mapping, Sequence

from atlassian import Jira  # type: ignore # noqa
from cogapp import Cog, CogUsageError  # type: ignore

import laskea
import laskea.config as cfg


def process(command: str, transaction_mode: str, paths: Sequence[str], options: Mapping[str, bool]) -> int:
    """Drive the lookup."""

    if command != 'update' or not paths:
        print('Usage: laskea update [--help] [-v] [-c config-path] [-n] [-i] source-files*md [other.md]')
        sys.exit(2)

    quiet = bool(options.get('quiet', ''))
    verbose = bool(options.get('verbose', ''))

    vector = [
        laskea.APP_ALIAS,
        '-P',
        '-c',
        f'--markers={laskea.BASE_MARKERS}',
        '-p',
        'from laskea import *',
    ]
    if transaction_mode == 'commit':
        vector.append('-r')
    if quiet:
        vector.append('--verbosity=0')

    cog = Cog()

    if laskea.DEBUG or verbose:
        cfg.report_context(command, transaction_mode, vector)

    for path in paths:
        single_vector = vector + [path]
        try:
            cog.callableMain(single_vector)
        except CogUsageError as err:
            print('CodeGen processing usage error:', file=sys.stderr)
            print(str(err))
            return 1
    return 0



import json
import pathlib

import pytest

import laskea
import laskea.config as cfg


def test_generate_template_command():
    json_string = cfg.generate_template()
    assert '"markers": "[[[fill ]]] [[[end]]]"' in json_string


def test_process_spoc_no_file(capsys):
    with pytest.raises(SystemExit):
        cfg.process('no-file', {})
    out, err = capsys.readouterr()
    assert 'Given configuration path is no file or empty' in err
    assert not out


def test_process_spoc_template_file(capsys):
    fixture_config = pathlib.Path('test', 'fixtures', 'basic', 'dot.laskea.json')
    assert cfg.process(str(fixture_config), {'quiet': laskea.QUIET, 'verbose': True}) is None
    out, err = capsys.readouterr()
    assert 'Configuration interface combined file, environment, and commandline values!' in err
    assert not out


def test_report_context(capsys):
    quiet_flag_restore = laskea.QUIET
    laskea.QUIET = False
    assert cfg.report_context(command='-', transaction_mode='+', vector=['42']) is None
    out, err = capsys.readouterr()
    assert not out
    lines = err.strip().split('\n')
    assert len(lines) == 21
    assert lines[:3] == ['Command: (-)', '- Transaction mode: (+)', 'Environment(variable values):']
    assert lines[11] == 'Effective(variable values):'
    for line in lines[3:11]:
        assert line.startswith(f'- {laskea.APP_ENV}_')
    assert lines[-1] == "- CallVector: (['42'])"
    laskea.QUIET = quiet_flag_restore


def test_report_context_quiet(capsys):
    quiet_flag_restore = laskea.QUIET
    laskea.QUIET = True
    assert cfg.report_context(command='-', transaction_mode='+', vector=['42']) is None
    out, err = capsys.readouterr()
    assert not out
    assert not err
    laskea.QUIET = quiet_flag_restore


def test_report_sources_of_effective_configuration(capsys):
    quiet_flag_restore = laskea.QUIET
    laskea.QUIET = False
    assert cfg.report_sources_of_effective_configuration(source_of={}, header='42') is None
    out, err = capsys.readouterr()
    assert not out
    lines = err.strip().split('\n')
    assert lines == ['42', '# --- BEGIN ---', '{}', '# --- E N D ---']
    laskea.QUIET = quiet_flag_restore


def test_report_sources_of_effective_configuration_quiet(capsys):
    quiet_flag_restore = laskea.QUIET
    laskea.QUIET = True
    assert cfg.report_sources_of_effective_configuration(source_of={}, header='42') is None
    out, err = capsys.readouterr()
    assert not out
    assert not err
    laskea.QUIET = quiet_flag_restore


def test_safe_report_configuration(capsys):
    quiet_flag_restore = laskea.QUIET
    laskea.QUIET = False
    assert cfg.safe_report_configuration(configuration={}, header='42') is None
    out, err = capsys.readouterr()
    assert not out
    lines = err.strip().split('\n')
    assert lines == ['42', '# --- BEGIN ---', '{}', '# --- E N D ---']
    laskea.QUIET = quiet_flag_restore


def test_safe_report_configuration_quiet(capsys):
    quiet_flag_restore = laskea.QUIET
    laskea.QUIET = True
    assert cfg.safe_report_configuration(configuration={}, header='42') is None
    out, err = capsys.readouterr()
    assert not out
    assert not err
    laskea.QUIET = quiet_flag_restore


def test_safe_report_configuration_no_leak(capsys):
    quiet_flag_restore = laskea.QUIET
    laskea.QUIET = False
    thing = {'remote': {'token': 'leak'}}
    gnith = {'remote': {'token': '*************'}}
    assert cfg.safe_report_configuration(configuration=thing, header='42') is None
    out, err = capsys.readouterr()
    assert not out
    lines = err.strip().split('\n')
    assert lines == ['42', '# --- BEGIN ---'] + json.dumps(gnith, indent=2).split('\n') + ['# --- E N D ---']
    laskea.QUIET = quiet_flag_restore


def test_create_and_report_effective_configuration(capsys):
    quiet_flag_restore = laskea.QUIET
    laskea.QUIET = False
    assert cfg.create_and_report_effective_configuration(header='42') is None
    out, err = capsys.readouterr()
    assert not out
    lines = err.strip().split('\n')
    assert len(lines) == 47
    assert lines[:2] == ['42', '# --- BEGIN ---']
    assert lines[-1] == '# --- E N D ---'
    print(lines)
    laskea.QUIET = quiet_flag_restore


def test_create_and_report_effective_configuration_quiet(capsys):
    quiet_flag_restore = laskea.QUIET
    laskea.QUIET = True
    assert cfg.create_and_report_effective_configuration(header='42') is None
    out, err = capsys.readouterr()
    assert not out
    assert not err
    laskea.QUIET = quiet_flag_restore


def test_load_configuration_empty(capsys):
    quiet_flag_restore = laskea.DEBUG
    base_markers_restore = laskea.BASE_MARKERS
    laskea.DEBUG = False
    assert cfg.load_configuration(configuration={}) == {}
    out, err = capsys.readouterr()
    assert not out
    assert err.strip() == 'Warning: Requested load from empty configuration'
    laskea.DEBUG = quiet_flag_restore
    laskea.BASE_MARKERS = base_markers_restore


def test_load_configuration_remote_token(capsys):
    quiet_flag_restore = laskea.DEBUG
    base_markers_restore = laskea.BASE_MARKERS
    laskea.DEBUG = False
    thing = {'remote': {'token': 'leak'}}
    gnith = {'remote_base_url': 'env', 'remote_token': 'env', 'remote_user': 'env'}
    assert cfg.load_configuration(configuration=thing) == gnith
    out, err = capsys.readouterr()
    assert not out
    assert not err
    laskea.DEBUG = quiet_flag_restore
    laskea.BASE_MARKERS = base_markers_restore



import pytest

import laskea
import laskea.cli as cli


def test_report_command(capsys):
    with pytest.raises(SystemExit):
        cli.report()
    out, err = capsys.readouterr()
    assert f'laskea : {laskea.__version__}' in out
    assert not err


def test_template_command(capsys):
    with pytest.raises(SystemExit):
        cli.app_template()
    out, err = capsys.readouterr()
    assert '"markers": "[[[fill ]]] [[[end]]]"' in out
    assert not err


def test_update_command(capsys):
    with pytest.raises(TypeError):
        cli.update(source=['not-present'])
    out, err = capsys.readouterr()
    assert not out
    assert not err



# -*- coding: utf-8 -*-
"""Tabulator style REST/JSON data proxy connector API for code generation."""
from typing import Mapping, no_type_check

import jmespath
import requests  # noqa
from urllib3.exceptions import InsecureRequestWarning  # noqa

import laskea

requests.packages.urllib3.disable_warnings(category=InsecureRequestWarning)  # type: ignore


@no_type_check
def parse_matrix(configuration: Mapping[str, object]) -> Mapping[str, object]:
    """DRY."""
    matrix = configuration['matrix']
    parsed = {
        'columns': tuple(key for key, _, _, _ in matrix),  # noqa
        'labels': tuple(label for _, label, _, _ in matrix),  # noqa
        'is_numeric': [key for key, _, is_num, _ in matrix if is_num],  # noqa
        'align': [lcr for _, _, _, lcr in matrix],  # noqa
    }

    parsed['out_round'] = tuple(
        (key, 1 if key in parsed['is_numeric'] else 0) for key, _ in zip(parsed['columns'], parsed['labels'])
    )
    parsed['out_format'] = tuple(
        (key, '5.1f' if key in parsed['is_numeric'] else '') for key, _ in zip(parsed['columns'], parsed['labels'])
    )
    parsed['column_fields'] = tuple((key, label) for key, label in zip(parsed['columns'], parsed['labels']))

    return parsed


@no_type_check
def tabulator_overview_table(configuration: Mapping[str, object]) -> str:
    """Because we can ... document later."""
    m = parse_matrix(configuration)
    over_view = f'{configuration["base_url"]}{configuration["path"]}'

    data = []
    data_version = ''
    for year in configuration['years']:  # noqa
        source = over_view.replace('$year$', str(year))
        r = requests.get(source, verify=configuration['verify_server_certificate'])  # noqa
        as_json = r.json()
        data_version = jmespath.search('data_version', as_json)
        for entry in jmespath.search('data[]', as_json):
            record = [entry[key] for key, _ in m['column_fields']]  # noqa
            for i, (l, f) in enumerate(m['out_format']):  # noqa
                if i and f:
                    if record[i] is not None:
                        try:
                            v = float(record[i])
                            record[i] = f'{round(v, m["out_round"][i][1]): {f}}'  # noqa
                        except ValueError:
                            pass
                    else:
                        record[i] = ''

            data.append(record)

    header_widths = [len(label) for _, label in m['column_fields']]  # noqa
    widths = header_widths[:]
    selection = data[:]
    for record in selection:
        for i, s in enumerate(record):
            if s is not None:
                widths[i] = max(widths[i], len(s))

    header_cells = [
        m['labels'][key].rjust(widths[key])  # noqa
        if m['align'][key] == 'R'  # noqa
        else (
            m['labels'][key].ljust(widths[key])  # noqa
            if m['align'][key] == 'L'  # noqa
            else m['labels'][key].center(widths[key])  # noqa
        )  # noqa
        for key in range(len(m['column_fields']))  # noqa
    ]
    header = f'| {" | ".join(header_cells)} |'

    separator_cells = ['-' * (widths[key] + 1) for key in range(len(m['column_fields']))]  # noqa
    separator = [
        f':{v}' if lcr == 'L' else (f'{v}:' if lcr == 'R' else f':{v[:-1]}:')
        for lcr, v in zip(m['align'], separator_cells)  # noqa
    ]
    separator_display = f'|{"|".join(separator)}|'

    rows = []
    for row in selection:
        rows.append(
            [
                str(v).rjust(widths[k])
                if m['align'][k] == 'R'  # noqa
                else (str(v).ljust(widths[k]) if m['align'][k] == 'L' else str(v).center(widths[k]))  # noqa
                for k, v in enumerate(row)
            ]
        )

    rows_display = [f'| {" | ".join(v for v in row)} |' for row in rows]

    summary = f'\n\nData version: {data_version}'
    the_table = '\n'.join([header] + [separator_display] + rows_display) + summary

    return the_table.replace('\r', '') if laskea.BASE_LF_ONLY else the_table


@no_type_check
def tabulator_kpi_table(configuration: Mapping[str, object], selected: str) -> str:
    """Because we can too ... document later."""
    m = parse_matrix(configuration)

    data = []
    data_version = ''
    for year in configuration['years']:  # noqa
        the_path = configuration['paths'][selected].replace('$year$', str(year))  # noqa
        source = f'{configuration["base_url"]}{the_path}'
        r = requests.get(source, configuration['verify_server_certificate'])  # noqa
        as_json = r.json()
        data_version = jmespath.search('data_version', as_json)
        for entry in jmespath.search('data[]', as_json):
            record = [entry[key] for key, _ in m['column_fields']]  # noqa
            for i, (l, f) in enumerate(m['out_format']):  # noqa
                if i and f and record[i] is not None:
                    record[i] = f'{round(record[i], m["out_round"][i][1]): {f}}'  # noqa
            data.append(record)

    header_widths = [len(label) for _, label in m['column_fields']]  # noqa
    widths = header_widths[:]
    selection = sorted(data, reverse=True)[:12]
    for record in selection:
        for i, s in enumerate(record):
            widths[i] = max(widths[i], len(s))

    header_cells = [
        m['labels'][key].rjust(widths[key]) if key else m['labels'][key].ljust(widths[key])  # noqa
        for key in range(len(m['column_fields']))  # noqa
    ]
    header = f'| {" | ".join(header_cells)} |'

    separator_cells = ['-' * (widths[key] + 1) for key in range(len(m['column_fields']))]  # noqa
    separator = f'|:{separator_cells[0]}|{":|".join(separator_cells[1:])}:|'

    rows = [f'| {" | ".join(str(v).rjust(widths[k]) for k, v in enumerate(line))} |' for line in selection]

    summary = f'\n\nData version: {data_version}'
    the_table = '\n'.join([header] + [separator] + rows) + summary

    return the_table.replace('\r', '') if laskea.BASE_LF_ONLY else the_table



"""Derive the license information and publish in docs."""
import functools
import json
import pathlib
import pkg_resources
import string
import subprocess  # nosec
from typing import List, Tuple

__all__ = ['dependency_tree_console_text', 'direct_dependencies_table', 'indirect_dependencies_table']

ENCODING = 'utf-8'
TP_PATH = pathlib.Path('docs', 'third-party')

TABLE_KEYS = (('Name', 'URL'), 'Version', 'License', 'Author', 'Description')
HEADER_LABELS = ('Name', 'Version', 'License', 'Author', 'Description (from packaging data)')
FALLBACK_URLS = {
    'typing-extensions': 'https://github.com/python/typing/blob/master/typing_extensions/README.rst',
}
TARGET = """\
__version__ = '$version$+parent.$revision$'\
"""


@functools.lru_cache()
def _fetch_direct_dependency_names():
    with pathlib.Path('requirements.txt').open(encoding=ENCODING) as requirements_txt:
        install_requires = [
            str(requirement)
            for requirement
            in pkg_resources.parse_requirements(requirements_txt)
        ]
    return install_requires


def _generate_dependency_information() -> None:
    """Use pip-licenses for creation of diverse databases and graphs."""
    install_requires = _fetch_direct_dependency_names()
    tokens = set(list(string.ascii_letters + '-_'))
    direct_names = [''.join(c for c in term if c in tokens) for term in install_requires]
    direct_vector = [
        'pip-licenses', '--format', 'json', '-p', *direct_names,
        '--with-authors', '--with-description', '--with-urls', '--with-license-file', '--with-notice-file',
        '--output-file', str(TP_PATH / 'direct-dependency-licenses.json')]
    noise = subprocess.run(direct_vector, capture_output=True, encoding=ENCODING, text=True, check=True).stdout.strip()  # nosec
    if not noise.startswith('created path: ') or not noise.endswith('direct-dependency-licenses.json'):
        raise RuntimeError(noise)

    indirect_names = [  # TODO(sthagen) these indirect deps may diverge ...
        'appdirs',
        'attrs',
        'cattrs',
        'certifi',
        'charset-normalizer',
        'click',
        'deprecated',
        'idna',
        'oauthlib',
        'requests',
        'requests-oauthlib',
        'six',
        'typing-extensions',
        'url-normalize',
        'urllib3',
        'wrapt',

    ]
    full_vector = [
        'pip-licenses', '--format', 'json', '-p', *direct_names, *indirect_names,
        '--with-authors', '--with-description', '--with-urls', '--with-license-file', '--with-notice-file',
        '--output-file', str(TP_PATH / 'all-dependency-licenses.json')]
    noise = subprocess.run(full_vector, capture_output=True, encoding=ENCODING, text=True, check=True).stdout.strip()  # nosec
    if not noise.startswith('created path: ') or not noise.endswith('all-dependency-licenses.json'):
        raise RuntimeError(noise)

    base_vector = ['pipdeptree', '--packages', ','.join(direct_names)]
    jobs = (
        (TP_PATH / 'package-dependency-tree.dot.txt', base_vector + ['--graph-output', 'dot']),
        (TP_PATH / 'package-dependency-tree.svg', base_vector + ['--graph-output', 'svg']),
        (TP_PATH / 'package-dependency-tree.json', base_vector + ['--json-tree', '--warn', 'silence']),
        (TP_PATH / 'package-dependency-tree.console.txt', base_vector + ['--warn', 'silence']),
    )
    for target, vector in jobs:
        plot = subprocess.run(vector, capture_output=True, encoding=ENCODING, text=True, check=True).stdout.strip()  # nosec
        target.write_text(plot, encoding=ENCODING)


@functools.lru_cache()
def _fetch_dependencies(direct_only: bool = True):
    db = 'direct-dependency-licenses.json' if direct_only else 'all-dependency-licenses.json'
    dep_json_path = pathlib.Path('docs', 'third-party') / db
    with open(dep_json_path, 'rt', encoding=ENCODING) as handle:
        data = json.load(handle)
    return data


def _markdown_table(table: List[Tuple[str, str, str, str, str]], header_labels=HEADER_LABELS) -> str:
    """Create the gfm table as string."""
    columns = header_labels
    col_wid = {key: len(key) for key in columns}
    for record in table:
        for key, cell in zip(columns, record):
            col_wid[key] = max(len(cell), col_wid[key])

    header_cells = [key.ljust(col_wid[key]) for key in columns]
    header = f'| {" | ".join(header_cells)} |'

    separator_cells = ['-' * (col_wid[key] + 1) for key in columns]
    separator = f'|:{"|:".join(separator_cells)}|'

    rows = [f'| {" | ".join(str(v).ljust(col_wid[k]) for k, v in zip(columns, line))} |' for line in table]

    return '\n'.join([header] + [separator] + rows)


def _extract_rows(data):
    rows = []
    for record in data:
        nam = record['Name']
        url = record.get('URL', '')
        if url == 'UNKNOWN':
            url = FALLBACK_URLS.get(nam, '')
        nam_e = f'[{nam}]({url})' if url else nam

        ver = record['Version']
        ver_sion = f'[{ver}](https://pypi.org/project/{nam}/{ver}/)'
        lic = record['License']
        aut = record['Author']
        des = record['Description']
        rows.append((nam_e, ver_sion, lic, aut, des))
    rows.sort()
    return rows


def direct_dependencies_table() -> None:
    """Fill in the data from the direct dependencies."""
    _generate_dependency_information()
    print(_markdown_table(_extract_rows(_fetch_dependencies(direct_only=True))))


def indirect_dependencies_table() -> None:
    """Fill in the data from the indirect dependencies."""
    direct_data = _fetch_dependencies(direct_only=True)
    direct_names = tuple(record['Name'] for record in direct_data)
    indirect_only_data = [rec for rec in _fetch_dependencies(direct_only=False) if rec['Name'] not in direct_names]
    print(_markdown_table(_extract_rows(indirect_only_data)))


def dependency_tree_console_text():
    """Fill in the pipdeptree console output minus any warnings."""
    console_tree = (TP_PATH / 'package-dependency-tree.console.txt').read_text(encoding=ENCODING).strip()
    fence = '````'
    print(f'{fence}console')
    print(console_tree)
    print(fence)



"""Calculate (Finnish: laskea) some parts."""
import os

# [[[fill git_describe()]]]
__version__ = '2022.9.21+parent.5187d766'
# [[[end]]] (checksum: 0e99719500ba864e5a56b7cdd35e9ddd)
__version_info__ = tuple(
    e if '-' not in e else e.split('-')[0] for part in __version__.split('+') for e in part.split('.') if e != 'parent'
)

APP_NAME = 'Calculate (Finnish: laskea) some parts.'
APP_ALIAS = 'laskea'
APP_ENV = 'LASKEA'
CACHE_EXPIRY_SECONDS = int(os.getenv(f'{APP_ENV}_CACHE_EXPIRY_SECONDS', '180'))
DEBUG = bool(os.getenv(f'{APP_ENV}_DEBUG', ''))
DRY_RUN = False
VERBOSE = bool(os.getenv(f'{APP_ENV}_VERBOSE', ''))
QUIET = False
STRICT = bool(os.getenv(f'{APP_ENV}_STRICT', ''))
ENCODING = 'utf-8'
ENCODING_ERRORS_POLICY = 'ignore'
DEFAULT_CONFIG_NAME = '.laskea.json'
OPEN_BRACKET = '['
CLOSE_BRACKET = ']'
DEFAULT_MARKERS = f'{OPEN_BRACKET * 3}fill {CLOSE_BRACKET * 3} {OPEN_BRACKET * 3}end{CLOSE_BRACKET * 3}'
DEFAULT_LF_ONLY = 'YES'
DEFAULT_JOIN_STRING = ' <br>'
BASE_MARKERS = os.getenv(f'{APP_ENV}_MARKERS', DEFAULT_MARKERS)
BASE_LF_ONLY = bool(os.getenv(f'{APP_ENV}_LF_ONLY', DEFAULT_LF_ONLY))
BASE_JOIN_STRING = os.getenv(f'{APP_ENV}_JOIN_STRING', DEFAULT_JOIN_STRING)
MASK_DISPLAY = '*' * 13
EXCEL = {'mbom': 'mbom.xlsm'}
TABULATOR = {
    'overview': {
        'base_url': 'https://example.com/metrics/',
        'path': '$year$/kpi-table-$year$.json',
        'years': [2022],
        'matrix': [
            ['section', 'Section', False, 'L'],
            ['name', 'Name', False, 'L'],
            ['unit', 'Unit', False, 'C'],
            ['all', 'ALL', True, 'R'],
            ['pr1', 'PR1', True, 'R'],
            ['pr2', 'PR2', True, 'R'],
            ['pr3', 'PR3', True, 'R'],
            ['description', 'Description', False, 'L'],
        ],
        'verify_server_certificate': False,
    },
    'metrics': {
        'base_url': 'https://example.com/metrics/',
        'paths': {
            'review_effectivity': '$year$/review_effectivity/kpi-review_effectivity-per_product-report-$year$.json',
            'sprint_effectivity': '$year$/sprint_effectivity/kpi-sprint_effectivity-per_product-report-$year$.json',
            'task_traceability': '$year$/task_traceability/kpi-task_traceability-per_product-report-$year$.json',
        },
        'years': [2021, 2022],
        'matrix': [
            ['month', 'Month', False, 'L'],
            ['all', 'ALL', True, 'R'],
            ['pr1', 'PR1', True, 'R'],
            ['pr2', 'PR2', True, 'R'],
            ['pr3', 'PR3', True, 'R'],
            ['trend_all', '±ALL', True, 'R'],
            ['trend_pr1', '±PR1', True, 'R'],
            ['trend_pr2', '±PR2', True, 'R'],
            ['trend_pr3', '±PR3', True, 'R'],
        ],
        'verify_server_certificate': False,
    },
}

from laskea.api.excel import mbom_table  # noqa
from laskea.api.jira import (  # noqa
    login,
    markdown_heading,
    markdown_list,
    markdown_table,
    parent_children_sections,
    query,
)
from laskea.embed import (  # noqa
    dl,
    h1,
    h2,
    h3,
    h4,
    h5,
    h6,
    kpi_table,
    mbom_table,
    metrics_table,
    ol,
    table,
    test_plans,
    ul,
)

__all__ = [
    'h1',
    'h2',
    'h3',
    'h4',
    'h5',
    'h6',
    'dl',
    'login',
    'markdown_heading',
    'markdown_table',
    'markdown_list',
    'query',
    'ol',
    'table',
    'ul',
    'kpi_table',
    'mbom_table',
    'metrics_table',
    'parent_children_sections',
    'test_plans',
]



# -*- coding: utf-8 -*-
"""JIRA proxy connector API for code generation."""
import copy
import json
import os
import sys
from typing import Iterable, Mapping, Sized, Union, no_type_check

import jmespath
from atlassian import Jira  # type: ignore # noqa
from requests.exceptions import HTTPError

import laskea

API_BASE_URL = 'https://example.com'

DEFAULT_COLUMN_FIELDS = ['Key', 'Summary', ['Priority', 'P'], 'Status', 'Custom Field Wun', 'Custom Field Other (CFO)']

WUN_ID = 'customfield_11501'
ANOTHER_ID = 'customfield_13901'
KNOWN_CI_FIELDS = {
    'key': ['key', 'key'],
    'summary': ['summary', 'fields.summary'],
    'priority': ['priority', 'fields.priority.name'],
    'status': ['status', 'fields.status.name'],
    'custom field name': [WUN_ID, f'fields.{WUN_ID}'],
    'custom field other': [ANOTHER_ID, f'fields.{ANOTHER_ID}[].value'],
}

BASE_USER = os.getenv(f'{laskea.APP_ENV}_USER', '')
BASE_PASS = os.getenv(f'{laskea.APP_ENV}_TOKEN', '')
BASE_URL = os.getenv(f'{laskea.APP_ENV}_BASE_URL', '')
BASE_IS_CLOUD = bool(os.getenv(f'{laskea.APP_ENV}_IS_CLOUD', ''))
BASE_COL_FIELDS = json.loads(os.getenv(f'{laskea.APP_ENV}_COL_FIELDS', json.dumps(DEFAULT_COLUMN_FIELDS)))
BASE_COL_MAPS = json.loads(os.getenv(f'{laskea.APP_ENV}_COL_MAPS', json.dumps(KNOWN_CI_FIELDS)))
BASE_JOIN_STRING = os.getenv(f'{laskea.APP_ENV}_JOIN_STRING', ' <br>')
BASE_LF_ONLY = bool(os.getenv(f'{laskea.APP_ENV}_LF_ONLY', 'YES'))
LF = '\n'


def mock(number: int) -> int:
    """Intermediate for starting the dev env in a valid state."""
    return number


def login(user: str = '', token: str = '', url: str = '', is_cloud: bool = False) -> Jira:  # nosec
    """LatAli"""
    if not user:
        user = BASE_USER
    if not token:
        token = BASE_PASS
    if not url:
        url = BASE_URL
    if not is_cloud:
        is_cloud = BASE_IS_CLOUD
    if not user or not token or not url:
        raise ValueError('User, Token, and URL are all required for login.')
    return Jira(url=url, username=user, password=token, cloud=is_cloud)


@no_type_check
def query(handle: Jira, jql_text: str, column_fields=None) -> dict:
    """EggLayingWoolMilkDear."""

    if not column_fields:
        column_fields = BASE_COL_FIELDS

    if not jql_text.strip():
        return {
            'jql_text': jql_text,
            'error': 'Empty JIRA Query Language text detected',
        }

    completed_column_fields = []
    for entry in column_fields:
        if isinstance(entry, str):
            candidate, concept, label = entry.lower(), entry, entry
        else:
            try:
                concept, label = entry
                candidate = concept.lower()
            except TypeError:
                return {
                    'jql_text': jql_text,
                    'column_fields': column_fields,
                    'parsed_columns': completed_column_fields,
                    'error': f'The column ({entry}) is neither a string nor a pair of (concept, label)',
                }

        for field in BASE_COL_MAPS.keys():
            if field in candidate:
                completed_column_fields.append(
                    {
                        'path': BASE_COL_MAPS[field][1],
                        'id': BASE_COL_MAPS[field][0],
                        'concept': concept,
                        'label': label,
                        'field': field,
                    }
                )

    if not completed_column_fields:
        return {
            'jql_text': jql_text,
            'column_fields': column_fields,
            'error': 'Completed column fields empty (no known fields?)',
        }

    try:
        issues = handle.jql(jql_text, limit=1000)
    except (HTTPError, RuntimeError) as err:
        return {
            'jql_text': jql_text,
            'column_fields': column_fields,
            'parsed_columns': completed_column_fields,
            'error': str(err),
        }

    pairs = [(col['label'], col['path']) for col in completed_column_fields]
    rows = [{label: jmespath.search(path, issue) or [''] for label, path in pairs} for issue in issues['issues']]
    return {
        'jql_text': jql_text,
        'column_fields': column_fields,
        'parsed_columns': completed_column_fields,
        'error': None,
        'rows': rows,
    }


@no_type_check
def markdown_table(
    handle: Jira, jql_text: str, column_fields=None, data: Mapping[str, Union[object, Iterable, Sized]] = None
) -> str:
    """Yes we can ... document later."""
    if data is None:
        data = query(handle, jql_text, column_fields)
    if data.get('error', ''):
        return json.dumps(data, indent=2)

    if not data['rows']:
        if laskea.STRICT:
            message = f'WARNING: received 0 results for JQL ({jql_text}) and table'
            if not laskea.DRY_RUN:
                print(message, file=sys.stderr)
            return message
        return ''

    table = copy.deepcopy(data['rows'])
    columns = list(table[0].keys())  # noqa
    col_wid = {key: len(key) for key in columns}
    for slot, record in enumerate(table):
        for key, cell in record.items():
            if key.lower() == 'key':
                table[slot][key] = f'[{cell}]({BASE_URL.strip("/")}/browse/{cell})'  # noqa
            if not isinstance(cell, str):
                table[slot][key] = BASE_JOIN_STRING.join(cell)  # noqa
            col_wid[key] = max(len(table[slot][key]), col_wid[key])  # noqa

    header_cells = [key.ljust(col_wid[key]) for key in columns]
    header = f'| {" | ".join(header_cells)} |'

    separator_cells = ['-' * (col_wid[key] + 1) for key in columns]
    separator = f'|:{"|:".join(separator_cells)}|'

    rows = [f'| {" | ".join(str(v).ljust(col_wid[k]) for k, v in line.items())} |' for line in table]
    issues = len(table)
    summary = f'\n\n{issues} issue{"" if issues == 1 else "s"}'
    the_table = '\n'.join([header] + [separator] + rows) + summary
    return the_table.replace('\r', '') if BASE_LF_ONLY else the_table


@no_type_check
def markdown_list(
    handle: Jira,
    jql_text: str,
    column_fields=None,
    list_type: str = 'ul',
    data: Mapping[str, Union[object, Iterable, Sized]] = None,
) -> str:
    """Yes we can ... document later."""
    if data is None:
        data = query(handle, jql_text, column_fields)
    if data.get('error', ''):
        return json.dumps(data, indent=2)

    if not data['rows']:
        if laskea.STRICT:
            message = f'WARNING: received 0 results for JQL ({jql_text}) and {list_type}'
            if not laskea.DRY_RUN:
                print(message, file=sys.stderr)
            return message
        return ''

    items = []
    for record in data['rows']:
        k, v = '', ''
        for key, cell in record.items():
            if key.lower() not in ('key', 'summary'):
                continue
            if key.lower() == 'key':
                k = f'[{cell}]({BASE_URL.strip("/")}/browse/{cell})'
            else:
                v = cell
        items.append((k, v))

    if list_type in ('ol', 'ul'):
        lt = '-' if list_type == 'ul' else '1.'  # implicit 'ol'
        xl = tuple(f'{lt} {key} - {summary}' for key, summary in items)
        the_list = '\n'.join(xl) + '\n'
        return the_list.replace('\r', '') if BASE_LF_ONLY else the_list
    if list_type == 'dl':
        # 'Term'
        # ':definition of term'
        #
        xl = tuple(f'{key}\n:{summary}\n' for key, summary in items)
        the_list = '\n'.join(xl) + '\n'
        return the_list.replace('\r', '') if BASE_LF_ONLY else the_list
    return f'Unexpected list type ({list_type}) in markdown_list not in ({("dl", "ol", "ul")})' + '\n'


@no_type_check
def markdown_heading(
    handle: Jira,
    jql_text: str,
    column_fields=None,
    level: int = 1,
    data: Mapping[str, Union[object, Iterable, Sized]] = None,
) -> str:
    """Yes we can ... document later."""
    if data is None:
        data = query(handle, jql_text, column_fields)
    if data.get('error', ''):
        return json.dumps(data, indent=2)

    if not data['rows']:
        if laskea.STRICT:
            message = f'WARNING: received 0 results instead of 1 for JQL ({jql_text}) and h{level}'
            if not laskea.DRY_RUN:
                print(message, file=sys.stderr)
            return message
        return ''

    items = []
    for record in data['rows']:
        k, v = '', ''
        for key, cell in record.items():
            if key.lower() not in ('key', 'summary'):
                continue
            if key.lower() == 'key':
                k = f'[{cell}]({BASE_URL.strip("/")}/browse/{cell})'
            else:
                v = cell
        items.append((k, v))
    received = len(items)
    if received != 1:
        if laskea.STRICT:
            message = f'WARNING: received {received} results instead of 1 for JQL ({jql_text}) and h{level}'
            if not laskea.DRY_RUN:
                print(message, file=sys.stderr)
            return message.replace('\r', '') if BASE_LF_ONLY else message
        return ''
    level_range = tuple(range(1, 6 + 1))
    if level in level_range:
        heading_token = '#' * level
        xl = tuple(f'{heading_token} {key} - {summary}' for key, summary in items)
        the_heading = '\n'.join(xl)
        return the_heading.replace('\r', '') if BASE_LF_ONLY else the_heading
    message = f'Unexpected level for heading ({level}) in markdown_heading not in ({level_range})'
    if not laskea.DRY_RUN:
        print(message, file=sys.stderr)
    return message


@no_type_check
def fetch_jql(handle: Jira, jql_text: str) -> dict:
    """Expose the JIRA result structure directly."""
    if not jql_text.strip():
        return {
            'jql_text': jql_text,
            'error': 'Empty JIRA Query Language text detected',
        }

    try:
        issues = handle.jql(jql_text, limit=1000)
    except (HTTPError, RuntimeError) as err:
        return {
            'jql_text': jql_text,
            'error': str(err),
        }

    return {
        'jql_text': jql_text,
        'data': issues,
        'error': None,
    }


@no_type_check
def parent_children_sections(
    handle: Jira,
    parent_jql: str,
    children_jql: str,
    parent_type_name: str,
    children_type_name: str,
    data: Mapping[str, Union[object, Iterable, Sized]] = None,
) -> str:
    """Create sub(sub)section level content representing the issue content from parent children filter results."""
    if data is None:
        data = {
            'parent_data': fetch_jql(handle, jql_text=parent_jql),
            'children_data': fetch_jql(handle, jql_text=children_jql),
        }
    if data['parent_data'].get('error', '') or data['children_data'].get('error', ''):
        return json.dumps(data, indent=2)

    doc = {}
    has_parents = {}

    for parent in data['parent_data']['data']['issues']:
        p_id = parent['id']
        p_key = parent['key']

        p_field = parent['fields']

        p_itn = p_field['issuetype']['name']
        # assert p_itn == parent_type_name
        p_sum = p_field['summary']
        p_des = p_field['description']  # None for parent type names
        p_epic = p_field['customfield_10006']  # TODO assuming here ...
        p_created = p_field['created']  # Textual timestamps like "2019-03-12T10:01:25.000+0100"
        p_updated = p_field['updated']

        doc[p_key] = {
            'id': p_id,
            'type': p_itn,
            'summary': p_sum,
            'description': p_des,
            'epic': p_epic,
            'created': p_created,
            'updated': p_updated,
            'children': {},
        }

        children = p_field['subtasks']

        for child in children:
            c_id = child['id']
            c_key = child['key']

            c_field = child['fields']

            c_itn = c_field['issuetype']['name']
            # assert c_itn == children_type_name
            c_sum = c_field['summary']

            doc[p_key]['children'][c_key] = {
                'id': c_id,
                'type': c_itn,
                'summary': c_sum,
                'description': None,
                'created': None,
                'updated': None,
            }
            if c_key not in has_parents:
                has_parents[c_key] = []
            has_parents[c_key].append(p_key)

    for child in data['children_data']['data']['issues']:
        c_id = child['id']
        c_key = child['key']

        c_field = child['fields']

        p_key = c_field['parent']['key']
        # assert p_key in has_parents[c_key]
        # p_itn = c_field['parent']['issuetype']['name']
        # assert p_itn == parent_type_name

        c_itn = c_field['issuetype']['name']
        # assert c_itn == children_type_name
        c_sum = c_field['summary']
        c_des = c_field['description']  # Table or list or notes for children type names
        c_created = c_field['created']
        c_updated = c_field['updated']

        doc[p_key]['children'][c_key]['description'] = c_des
        doc[p_key]['children'][c_key]['created'] = c_created
        doc[p_key]['children'][c_key]['updated'] = c_updated

    return doc_to_markdown(doc, parent_type_name, children_type_name)


@no_type_check
def doc_to_markdown(doc, parent_type_name: str, children_type_name: str) -> str:  # noqa
    """Transform the document content to markdown."""
    md = []
    for p_tree in doc.values():
        p_head = f'## {p_tree["summary"]}'.strip().strip(LF)
        c_count = len(p_tree['children'])
        c_type_disp = f'{children_type_name}{"" if c_count == 1 else "s"}'
        p_para = f'The {p_tree["type"]} consists of {c_count} {c_type_disp}'.strip(LF)

        c_parts = []
        double_pipe, ast_pipe, pipe_ast = '||', '|*', '*|'
        nbsp = '&nbsp;'
        for c_data in p_tree['children'].values():
            c_head = f'### {c_data["summary"]}'.strip().strip(LF)
            c_in = list(c_data['description'].replace(nbsp, ' ').strip().split(LF))
            c_out = []
            for line in c_in:
                if line.startswith(double_pipe) or line.startswith(ast_pipe):
                    # patch confluence markdown like table heads ...
                    line_s = line.strip()
                    extra_line = ''
                    if line_s.startswith(double_pipe) and line_s.endswith(double_pipe):
                        line = line.replace(double_pipe, '|')
                        extra_line = (
                            ''.join(c if c == '|' else '-' for c in line).replace('|-', '|:').replace('-|', ' |')
                        )
                    elif line_s.startswith(ast_pipe) and line_s.endswith(pipe_ast):
                        extra_line = (
                            ''.join(c if c == '|' else '-' for c in line).replace('|-', '|:').replace('-|', ' |')
                        )
                    c_out.append(line)
                    if extra_line:
                        c_out.append(extra_line)
                else:
                    c_out.append(line)

            c_parts.extend([LF, c_head, LF, *c_out])

        md.extend([LF, p_head, LF, p_para])
        md.extend(c_parts)

    md.append(LF)
    return LF.join(md).replace(LF + LF, LF)



"""Configuration API for laskea."""
import copy
import json
import os
import pathlib
import sys
from typing import Dict, List, Mapping, Tuple, no_type_check

import jmespath

import laskea
import laskea.api.jira as api

TEMPLATE_EXAMPLE = """\
{
  "table": {
    "column": {
      "fields": [
        "Key",
        "Summary",
        "Custom Field Name",
        "Custom Field Other"
      ],
      "field_map": {
        "key": [
          "key",
          "key"
        ],
        "summary": [
          "summary",
          "fields.summary"
        ],
        "custom field name": [
          "customfield_11501",
          "fields.customfield_11501"
        ],
        "custom field other": [
          "customfield_13901",
          "fields.customfield_13901[].value"
        ]
      },
      "lf_only": true,
      "join_string": " <br>"
    }
  },
  "remote": {
    "is_cloud": false,
    "user": "",
    "token": "",
    "base_url": "https://remote-jira-instance.example.com/"
  },
  "local": {
    "markers": "[[[fill ]]] [[[end]]]",
    "quiet": false,
    "verbose": false,
    "strict": false
  },
  "excel": {
    "mbom": "mbom.xlsm"
  },
  "tabulator": {
    "overview": {
        "base_url": "https://example.com/metrics/",
        "path": "$year$/kpi-table-$year$.json",
        "years": [2022],
        "matrix": [
            ["section", "Section", False, "L"],
            ["name", "Name", False, "L"],
            ["unit", "Unit", False, "C"],
            ["all", "ALL", True, "R"],
            ["pr1", "PR1", True, "R"],
            ["pr2", "PR2", True, "R"],
            ["pr3", "PR3", True, "R"],
            ["description", "Description", False, "L"]
        ]
    },
    "metrics": {
        "base_url": "https://example.com/metrics/",
        "paths": {
            "review_effectivity": "$year$/review_effectivity/kpi-review_effectivity-per_product-report-$year$.json",
            "sprint_effectivity": "$year$/sprint_effectivity/kpi-sprint_effectivity-per_product-report-$year$.json",
            "task_traceability": "$year$/task_traceability/kpi-task_traceability-per_product-report-$year$.json",
        },
        "years": [2021, 2022],
        "matrix": [
            ["month", "Month", False, "L"],
            ["all", "ALL", True, "R"],
            ["pr1", "PR1", True, "R"],
            ["pr2", "PR2", True, "R"],
            ["pr3", "PR3", True, "R"],
            ["trend_all", "±ALL", True, "R"],
            ["trend_pr1", "±PR1", True, "R"],
            ["trend_pr2", "±PR2", True, "R"],
            ["trend_pr3", "±PR3", True, "R"]
        ]
    }
  }
}
"""


def generate_template() -> str:
    """Return template of a well-formed JSON configuration."""
    return TEMPLATE_EXAMPLE


@no_type_check
def load_configuration(configuration: Dict[str, object]) -> Dict[str, str]:
    """LaterAlligator."""
    if not configuration:
        print('Warning: Requested load from empty configuration', file=sys.stderr)
        return {}

    source_of = {}

    column_fields = jmespath.search('table.column.fields[]', configuration)
    if column_fields:
        source_of['column_fields'] = 'config'
        api.BASE_COL_FIELDS = copy.deepcopy(column_fields)
    column_fields = os.getenv(f'{laskea.APP_ENV}_COL_FIELDS', '')
    if column_fields:
        source_of['column_fields'] = 'env'
        api.BASE_COL_FIELDS = json.loads(column_fields)

    field_map = jmespath.search('table.column.field_map', configuration)
    if field_map:
        source_of['field_map'] = 'config'
        api.BASE_COL_MAPS = copy.deepcopy(field_map)
    field_map = os.getenv(f'{laskea.APP_ENV}_COL_MAPS', '')
    if field_map:
        source_of['field_map'] = 'env'
        api.BASE_COL_MAPS = json.loads(field_map)

    lf_only = jmespath.search('table.column.lf_only', configuration)
    if lf_only:
        source_of['lf_only'] = 'config'
        api.BASE_LF_ONLY = lf_only
    lf_only = os.getenv(f'{laskea.APP_ENV}_LF_ONLY', '')
    if lf_only:
        source_of['lf_only'] = 'env'
        api.BASE_LF_ONLY = lf_only

    join_string = jmespath.search('table.column.join_string', configuration)
    if join_string:
        source_of['join_string'] = 'config'
        api.BASE_JOIN_STRING = join_string
    join_string = os.getenv(f'{laskea.APP_ENV}_JOIN_STRING', '')
    if join_string:
        source_of['join_string'] = 'env'
        api.BASE_JOIN_STRING = join_string

    remote_user = jmespath.search('remote.user', configuration)
    if remote_user:
        source_of['remote_user'] = 'config'
        api.BASE_USER = remote_user
    remote_user = os.getenv(f'{laskea.APP_ENV}_USER', '')
    if remote_user:
        source_of['remote_user'] = 'env'
        api.BASE_USER = remote_user

    remote_token = jmespath.search('remote.token', configuration)
    if remote_token:
        source_of['remote_token'] = 'config'  # nosec
        api.BASE_TOKEN = remote_token
    remote_token = os.getenv(f'{laskea.APP_ENV}_TOKEN', '')
    if remote_token:
        source_of['remote_token'] = 'env'  # nosec
        api.BASE_TOKEN = remote_token

    remote_base_url = jmespath.search('remote.base_url', configuration)
    if remote_base_url:
        source_of['remote_base_url'] = 'config'
        api.BASE_URL = remote_base_url
    remote_base_url = os.getenv(f'{laskea.APP_ENV}_BASE_URL', '')
    if remote_base_url:
        source_of['remote_base_url'] = 'env'
        api.BASE_URL = remote_base_url

    local_markers = jmespath.search('local.markers', configuration)
    if local_markers:
        source_of['local_markers'] = 'config'
        laskea.BASE_MARKERS = local_markers
    local_markers = os.getenv(f'{laskea.APP_ENV}_MARKERS', '')
    if local_markers:
        source_of['local_markers'] = 'env'
        laskea.BASE_MARKERS = local_markers

    verbose = bool(jmespath.search('local.verbose', configuration))
    if verbose:
        source_of['verbose'] = 'config'
        laskea.DEBUG = verbose
    verbose = bool(os.getenv(f'{laskea.APP_ENV}_DEBUG', ''))
    if verbose:
        source_of['verbose'] = 'env'
        laskea.DEBUG = verbose

    is_cloud = bool(jmespath.search('remote.is_cloud', configuration))
    if is_cloud:
        source_of['is_cloud'] = 'config'
        laskea.IS_CLOUD = is_cloud
    is_cloud = bool(os.getenv(f'{laskea.APP_ENV}_IS_CLOUD', ''))
    if is_cloud:
        source_of['is_cloud'] = 'env'
        laskea.IS_CLOUD = is_cloud

    strict = bool(jmespath.search('local.strict', configuration))
    if strict:
        source_of['strict'] = 'config'
        laskea.STRICT = strict
    strict = bool(os.getenv(f'{laskea.APP_ENV}_STRICT', ''))
    if strict:
        source_of['strict'] = 'env'
        laskea.STRICT = strict

    quiet = bool(jmespath.search('local.quiet', configuration))
    if quiet:
        source_of['quiet'] = 'config'
        laskea.QUIET = quiet
        if source_of['verbose'] == 'config':
            laskea.DEBUG = quiet
    quiet = bool(os.getenv(f'{laskea.APP_ENV}_QUIET', ''))
    if quiet:
        source_of['quiet'] = 'env'
        laskea.QUIET = quiet
        source_of['verbose'] = 'env'
        laskea.DEBUG = quiet

    if 'tabulator' in configuration:
        laskea.TABULATOR = copy.deepcopy(configuration['tabulator'])

    return source_of


@no_type_check
def discover_configuration(conf: str) -> Tuple[Dict[str, object], str]:
    """Try to retrieve the configuration following the "(explicit, local, parents, home)
    first wun wins" strategy."""
    configuration = None
    if conf:
        cp = pathlib.Path(conf)
        if not cp.is_file() or not cp.stat().st_size:
            print('Given configuration path is no file or empty', file=sys.stderr)
            sys.exit(2)
        if not laskea.QUIET:
            print(f'Reading configuration file {cp} as requested...', file=sys.stderr)
        with cp.open(encoding=laskea.ENCODING) as handle:
            configuration = json.load(handle)
    else:
        cn = laskea.DEFAULT_CONFIG_NAME
        cwd = pathlib.Path.cwd().resolve()
        for pp in (cwd, *cwd.parents):
            cp = pp / cn
            if cp.is_file() and cp.stat().st_size:
                if not laskea.QUIET:
                    print(f'Reading from discovered configuration path {cp}', file=sys.stderr)
                with cp.open() as handle:
                    configuration = json.load(handle)
                return configuration, str(cp)

        cp = pathlib.Path.home() / laskea.DEFAULT_CONFIG_NAME
        if cp.is_file() and cp.stat().st_size:
            if not laskea.QUIET:
                print(
                    f'Reading configuration file {cp} from home directory at {pathlib.Path.home()} ...',
                    file=sys.stderr,
                )
            with cp.open() as handle:
                configuration = json.load(handle)
            return configuration, str(cp)

        if not laskea.QUIET:
            print(
                f'User home configuration path to {cp} is no file or empty - ignoring configuration data',
                file=sys.stderr,
            )

    return configuration, str(cp)


@no_type_check
def report_context(command: str, transaction_mode: str, vector: List[str]) -> None:
    """DRY."""
    if laskea.QUIET:
        return
    print(f'Command: ({command})', file=sys.stderr)
    print(f'- Transaction mode: ({transaction_mode})', file=sys.stderr)
    print('Environment(variable values):', file=sys.stderr)
    app_env_user = f'{laskea.APP_ENV}_USER'
    app_env_token = f'{laskea.APP_ENV}_TOKEN'
    app_env_base_url = f'{laskea.APP_ENV}_BASE_URL'
    app_env_col_fields = f'{laskea.APP_ENV}_COL_FIELDS'
    app_env_col_maps = f'{laskea.APP_ENV}_COL_MAPS'
    app_env_markers = f'{laskea.APP_ENV}_MARKERS'
    app_env_lf_only = f'{laskea.APP_ENV}_LF_ONLY'
    app_env_join_string = f'{laskea.APP_ENV}_JOIN_STRING'
    empty = ''
    print(f'- {laskea.APP_ENV}_USER: ({os.getenv(app_env_user, empty)})', file=sys.stderr)
    print(
        f'- {laskea.APP_ENV}_TOKEN: ({laskea.MASK_DISPLAY if len(os.getenv(app_env_token, empty)) else empty})',
        file=sys.stderr,
    )
    print(f'- {laskea.APP_ENV}_BASE_URL: ({os.getenv(app_env_base_url, empty)})', file=sys.stderr)
    print(f'- {laskea.APP_ENV}_COL_FIELDS: ({os.getenv(app_env_col_fields, empty)})', file=sys.stderr)
    print(f'- {laskea.APP_ENV}_COL_MAPS: ({os.getenv(app_env_col_maps, empty)})', file=sys.stderr)
    print(f'- {laskea.APP_ENV}_MARKERS: ({os.getenv(app_env_markers, empty)})', file=sys.stderr)
    print(f'- {laskea.APP_ENV}_LF_ONLY: ({os.getenv(app_env_lf_only, empty)})', file=sys.stderr)
    print(f'- {laskea.APP_ENV}_JOIN_STRING: ({os.getenv(app_env_join_string, empty)})', file=sys.stderr)
    print('Effective(variable values):', file=sys.stderr)
    print(f'- RemoteUser: ({api.BASE_USER})', file=sys.stderr)
    print(f'- RemoteToken: ({"*" * len(api.BASE_PASS)})', file=sys.stderr)
    print(f'- RemoteBaseURL: ({api.BASE_URL})', file=sys.stderr)
    print(f'- ColumnFields(table): ({api.BASE_COL_FIELDS})', file=sys.stderr)
    print(f'- ColumnMaps(remote->table): ({api.BASE_COL_MAPS})', file=sys.stderr)
    print(f'- Markers(pattern): ({laskea.BASE_MARKERS})', file=sys.stderr)
    print(f'- lf_only: ({laskea.BASE_LF_ONLY})', file=sys.stderr)
    print(f'- join_string: ({laskea.BASE_JOIN_STRING})', file=sys.stderr)
    print(f'- CallVector: ({vector})', file=sys.stderr)


@no_type_check
def report_sources_of_effective_configuration(source_of: Dict[str, str], header: str) -> None:
    """DRY."""
    if laskea.QUIET:
        return
    print(header, file=sys.stderr)
    print('# --- BEGIN ---', file=sys.stderr)
    print(json.dumps(source_of, indent=2), file=sys.stderr)
    print('# --- E N D ---', file=sys.stderr)


@no_type_check
def safe_report_configuration(configuration: Dict[str, object], header: str) -> None:
    """DRY."""
    if laskea.QUIET:
        return
    print(header, file=sys.stderr)
    print('# --- BEGIN ---', file=sys.stderr)
    fake_configuration = copy.deepcopy(configuration)
    if jmespath.search('remote.token', fake_configuration):
        fake_configuration['remote']['token'] = laskea.MASK_DISPLAY  # noqa
    print(json.dumps(fake_configuration, indent=2), file=sys.stderr)
    print('# --- E N D ---', file=sys.stderr)


@no_type_check
def create_and_report_effective_configuration(header: str) -> None:
    """DRY."""
    if laskea.QUIET:
        return
    effective = {
        'table': {
            'column': {
                'fields': copy.deepcopy(api.BASE_COL_FIELDS),
                'field_map': copy.deepcopy(api.BASE_COL_MAPS),
                'lf_only': api.BASE_LF_ONLY,
                'join_string': api.BASE_JOIN_STRING,
            },
        },
        'remote': {
            'is_cloud': api.BASE_IS_CLOUD,
            'user': api.BASE_USER,
            'token': '',
            'base_url': api.BASE_URL,
        },
        'local': {
            'markers': laskea.BASE_MARKERS,
            'quiet': laskea.QUIET,
            'verbose': laskea.DEBUG,
            'strict': laskea.STRICT,
        },
    }
    safe_report_configuration(effective, header)


def process(conf: str, options: Mapping[str, bool]) -> None:
    """SPOC."""
    configuration, cp = discover_configuration(conf)

    verbose = bool(options.get('verbose', ''))
    if configuration is not None:
        if laskea.DEBUG or verbose:
            safe_report_configuration(configuration, f'Loaded configuration from {cp}:')

        source_of = load_configuration(configuration)

        if laskea.DEBUG or verbose:
            report_sources_of_effective_configuration(source_of, f'Configuration source after loading from {cp}:')

        if not laskea.QUIET:
            print('Configuration interface combined file, environment, and commandline values!', file=sys.stderr)

        create_and_report_effective_configuration(
            f'Effective configuration combining {cp}, environment variables, and defaults:'
        )

    print(
        f'INFO: Upstream JIRA instance is addressed per {"cloud" if api.BASE_IS_CLOUD else "server"} rules',
        file=sys.stderr,
    )



import hashlib
import pathlib
from typing import no_type_check

from openpyxl import load_workbook  # type: ignore

from laskea import BASE_LF_ONLY

CHUNK_SIZE = 2 << 15


def hash_file(path: pathlib.Path) -> str:
    """Return the SHA512 hex digest of the data from file."""
    hash = hashlib.sha512()
    with open(path, 'rb') as handle:
        while chunk := handle.read(CHUNK_SIZE):
            hash.update(chunk)
    return hash.hexdigest()


@no_type_check
def mbom_table(filename: str):
    """Import MBOM per convention from office table file."""
    wb = load_workbook(filename=filename, data_only=True, read_only=True)
    ws = wb.worksheets[0]

    table_headers = [ws[address].value for address in ('A1', 'B1', 'C1', 'D1')]

    table_caption_data = [ws[address].value for address in ('A2', 'B2', 'C2', 'D2')]

    max_row = 1000
    table_rows = []
    for row in range(3, max_row):
        data = [ws[address].value for address in (f'A{row}', f'B{row}', f'C{row}', f'D{row}')]
        if [x for x in data if x]:
            table_rows.append(data)

    # print(table_headers)
    # ['Level', 'P/N', 'Item Name', 'SW Version']

    # print(table_caption_data)
    # [0, '123', 'SW, PRODUCT, CUST', None]

    # for row in table_rows:
    #     print(row)

    # [1, '124', 'THAT, DEF, LINUX', 'v1.2']
    # [1, '125', 'THAT, EFG, ABA8000, LINUX', 'Rev.24']
    # [1, '126', 'THAT, ABC, CUST', 'v7_R2427']
    # [2, '127', 'THAT, MNMNM, LINUX', 'v6.204.70']
    # [2, '128', 'THAT, WHAT, XXXX4, LINUX', 'v4.2']
    # [3, '129', 'SIMULATION, ABC v7.0', 'v7.0']
    # [4, '130', 'ORGA v7.0 IJK', 'v7.0']
    # [3, '222', 'ABC Config XML - CUST', 'v7.0']
    # [3, '223', 'ABC Default Trip XML', 'v7.0']
    # [3, '224', 'ABC Default Patching XML - LD Smoke', 'v7.0']
    # [3, '227', 'ABC Default EFK Config Files - MNO12', 'v6.204.70']
    # [2, '132', 'LIBRARY, FUNNY HEART, 1.0', 'v1.0']
    # [2, '131', 'XYZ RUNTIME, LINUX', 'v3.3.0']
    # [1, '141', 'THAT, UVW', 'Core 2.0.1, Adapter 2.0.4']

    header_widths = [len(label) for label in table_headers]  # noqa
    widths = header_widths[:]
    selection = table_rows[:]
    for record in selection:
        for i, s in enumerate(record):
            if s is not None:
                widths[i] = max(widths[i], len(str(s)))

    header_cells = [table_headers[key].ljust(widths[key]) for key in range(len(table_headers))]  # noqa  # noqa
    header = f'| {" | ".join(header_cells)} |'

    separator_cells = ['-' * (widths[key] + 1) for key in range(len(table_headers))]  # noqa
    separator = [f':{v}' for v in separator_cells]
    separator_display = f'|{"|".join(separator)}|'

    dows = []
    for dow in selection:
        dows.append([str(v).ljust(widths[k]) for k, v in enumerate(dow)])

    rows_display = [f'| {" | ".join(v for v in row)} |' for row in dows]

    semantics = f'<!-- anchor: {tuple(str(e) if e is not None else "" for e in table_caption_data)}-->'
    source = f'<!-- source: {pathlib.Path(filename)}-->'
    checksum = f'<!-- s-hash: sha512:{hash_file(pathlib.Path(filename))}-->'
    the_table = '\n'.join([semantics] + [header] + [separator_display] + rows_display + [source] + [checksum])

    return the_table.replace('\r', '') if BASE_LF_ONLY else the_table



"""Report environment to support resolution of user issues."""
import sys
from typing import List, no_type_check

import scooby  # type: ignore


@no_type_check
def report(shallow: bool = False) -> str:
    """Return either text options for the user to report her env or the report of the environment for support."""
    deep = True
    try:
        import pkg_resources  # noqa
    except (ImportError, ModuleNotFoundError):
        if shallow:
            deep = False
            sys.stdout.write('\nInfo: No setuptools module found and shallow reporting requested. Continuing ...')
        else:
            message = (
                '\n  Report requires setuptools (to identify the version of atlassian-python-api).'
                '\n  You may either add --shallow or install the module via `pip install setuptools`\n\n'
            )
            return message

    if deep:
        import atlassian  # type: ignore # noqa

        packages = pkg_resources.working_set  # noqa
        monkey_atl = [p.version for p in packages if p.project_name == 'atlassian-python-api'][0]  # noqa
        atlassian.__version__ = monkey_atl

    class Report(scooby.Report):  # type: ignore
        def __init__(self, additional=None, ncol=3, text_width=80, sort=False):
            """Initiate a scooby.Report instance."""

            # Mandatory packages.
            core = [
                'laskea',
                'atlassian',  # has version in VERSION text file in package info only?
                'cogapp.cogapp',
                'jmespath',
                'pydantic',
                'requests_cache',
                'scooby',
                'typer',
            ]

            # Optional packages.
            optional: List[str] = []

            scooby.Report.__init__(
                self, additional=additional, core=core, optional=optional, ncol=ncol, text_width=text_width, sort=sort
            )

    return str(Report()) + '\n'



"""Calculate (Finnish: laskea) some parts - embeddings."""
from typing import Dict, Union, no_type_check

from atlassian import Jira  # type: ignore # noqa

import laskea
import laskea.api.excel as exc
import laskea.api.jira as api
import laskea.api.tabulator as tab

DB: Dict[str, Union[None, Jira]] = {'handle': None}


@no_type_check
def mbom_table(configuration=None) -> None:
    """Public document interface."""
    if configuration is None:
        configuration = laskea.EXCEL['mbom']

    print(exc.mbom_table(configuration))


@no_type_check
def metrics_table(configuration=None) -> None:
    """Public document interface."""
    if configuration is None:
        configuration = laskea.TABULATOR['overview']

    print(tab.tabulator_overview_table(configuration))


@no_type_check
def kpi_table(selected, configuration=None) -> None:
    """Public document interface."""
    if configuration is None:
        configuration = laskea.TABULATOR['metrics']

    print(tab.tabulator_kpi_table(configuration, selected))


@no_type_check
def test_plans(
    parent_jql: str = '',
    children_jql: str = '',
    parent_type: str = 'Test Plan',
    children_type: str = 'Test Case',
    data=None,
) -> None:
    """Public document interface for the sub(sub)section document part generation from JIRA parents with children."""
    if data is None:
        if not DB.get('handle', None):
            DB['handle'] = api.login()

        print(api.parent_children_sections(DB['handle'], parent_jql, children_jql, parent_type, children_type))
    else:
        print(api.parent_children_sections(DB['handle'], parent_jql, children_jql, parent_type, children_type, data))


@no_type_check
def table(query_text: str = '', data=None) -> None:
    """Public document interface."""
    if data is None:
        if not DB.get('handle', None):
            DB['handle'] = api.login()

        print(api.markdown_table(DB['handle'], query_text))
    else:
        print(api.markdown_table(DB['handle'], query_text, data=data))


@no_type_check
def dl(query_text: str = '', data=None) -> None:
    """Public document interface for definition list."""
    if data is None:
        if not DB.get('handle', None):
            DB['handle'] = api.login()

        print(api.markdown_list(DB['handle'], query_text, list_type='dl'))
    else:
        print(api.markdown_list(DB['handle'], query_text, list_type='dl', data=data))


@no_type_check
def ol(query_text: str = '', data=None) -> None:
    """Public document interface for ordered list."""
    if data is None:
        if not DB.get('handle', None):
            DB['handle'] = api.login()

        print(api.markdown_list(DB['handle'], query_text, list_type='ol'))
    else:
        print(api.markdown_list(DB['handle'], query_text, list_type='ol', data=data))


@no_type_check
def ul(query_text: str = '', data=None) -> None:
    """Public document interface for unordered list."""
    if data is None:
        if not DB.get('handle', None):
            DB['handle'] = api.login()

        print(api.markdown_list(DB['handle'], query_text, list_type='ul'))
    else:
        print(api.markdown_list(DB['handle'], query_text, list_type='ul', data=data))


@no_type_check
def hx(level: int, query_text: str = '', data=None) -> None:
    """Public document interface for headings 1 through 6."""
    if data is None:
        if not DB.get('handle', None):
            DB['handle'] = api.login()

        print(api.markdown_heading(DB['handle'], query_text, level=level))
    else:
        print(api.markdown_heading(api.Jira(''), query_text, level=level, data=data))


@no_type_check
def h1(query_text: str = '', data=None) -> None:
    """Public document interface for heading 1."""
    return hx(1, query_text, data)


@no_type_check
def h2(query_text: str = '', data=None) -> None:
    """Public document interface for heading 2."""
    return hx(2, query_text, data)


@no_type_check
def h3(query_text: str = '', data=None) -> None:
    """Public document interface for heading 3."""
    return hx(3, query_text, data)


@no_type_check
def h4(query_text: str = '', data=None) -> None:
    """Public document interface for heading 4."""
    return hx(4, query_text, data)


@no_type_check
def h5(query_text: str = '', data=None) -> None:
    """Public document interface for heading 5."""
    return hx(5, query_text, data)


@no_type_check
def h6(query_text: str = '', data=None) -> None:
    """Public document interface for heading 6."""
    return hx(6, query_text, data)
